import {
  ErrorOverlay
} from "./chunk-NFE6IAQR.js";
import {
  SX_PROP_HELPER_TEXT
} from "./chunk-Z6S6EGNW.js";

// src/Chart.tsx
import * as React from "react";
import { createComponent } from "@mui/toolpad-core";
import { Container, Skeleton } from "@mui/material";
import {
  XAxis,
  YAxis,
  Tooltip,
  Legend,
  Line,
  ResponsiveContainer,
  CartesianGrid,
  ComposedChart,
  Bar,
  Area,
  Scatter
} from "recharts";
import { errorFrom } from "@mui/toolpad-utils/errors";
var CHART_DATA_SERIES_KINDS = ["line", "bar", "area", "scatter"];
function getBarChartDataSeriesNormalizedYKey(dataSeries, index) {
  return `${dataSeries.label}-${dataSeries.yKey}-${index}`;
}
function Chart({ data = [], loading, error, height, sx }) {
  const xValues = React.useMemo(
    () => data.flatMap((dataSeries) => {
      if (!dataSeries.xKey || !dataSeries.data) {
        return [];
      }
      return dataSeries.data.map((dataSeriesPoint) => dataSeriesPoint[dataSeries.xKey]);
    }).filter((value, index, array) => array.indexOf(value) === index).sort(
      (a, b) => typeof a === "number" && typeof b === "number" ? a - b : 0
    ),
    [data]
  );
  const barChartData = React.useMemo(() => {
    return xValues.map((xValue) => {
      const yValues = data.reduce((acc, dataSeries, index) => {
        if (dataSeries.kind !== "bar" || !dataSeries.xKey || !dataSeries.yKey) {
          return acc;
        }
        const point = (dataSeries.data || []).find(
          (dataSeriesPoint) => dataSeriesPoint[dataSeries.xKey] === xValue
        );
        return {
          ...acc,
          [getBarChartDataSeriesNormalizedYKey(dataSeries, index)]: point ? point[dataSeries.yKey] : 0
        };
      }, {});
      return {
        x: xValue,
        ...yValues
      };
    });
  }, [data, xValues]);
  const hasNonNumberXValues = xValues.some((xValue) => typeof xValue !== "number");
  const displayError = error ? errorFrom(error) : null;
  const isDataVisible = !loading && !displayError;
  return /* @__PURE__ */ React.createElement(Container, { disableGutters: true, sx: { ...sx, position: "relative" }, "aria-busy": loading }, /* @__PURE__ */ React.createElement(ResponsiveContainer, { width: "100%", height }, /* @__PURE__ */ React.createElement(ComposedChart, { data: barChartData, margin: { top: 20, right: 80 } }, isDataVisible ? /* @__PURE__ */ React.createElement(React.Fragment, null, /* @__PURE__ */ React.createElement(CartesianGrid, { strokeDasharray: "3 3" }), /* @__PURE__ */ React.createElement(
    XAxis,
    {
      dataKey: "x",
      type: hasNonNumberXValues ? "category" : "number",
      allowDuplicatedCategory: false,
      domain: hasNonNumberXValues ? void 0 : [Math.min(...xValues), Math.max(...xValues)]
    }
  ), /* @__PURE__ */ React.createElement(YAxis, { width: 80 }), /* @__PURE__ */ React.createElement(Tooltip, null), /* @__PURE__ */ React.createElement(Legend, null), data.map((dataSeries, index) => {
    if (!dataSeries.data || dataSeries.data.length === 0 || !dataSeries.xKey || !dataSeries.yKey) {
      return null;
    }
    const key = `${dataSeries.label}-${index}`;
    const normalizedData = dataSeries.data.map((dataSeriesPoint) => ({
      x: dataSeriesPoint[dataSeries.xKey],
      [dataSeries.yKey]: dataSeriesPoint[dataSeries.yKey]
    })).sort(
      (a, b) => typeof a[dataSeries.xKey] === "number" && typeof b[dataSeries.xKey] === "number" ? a[dataSeries.xKey] - b[dataSeries.xKey] : 0
    );
    switch (dataSeries.kind) {
      case "bar":
        return /* @__PURE__ */ React.createElement(
          Bar,
          {
            key,
            dataKey: getBarChartDataSeriesNormalizedYKey(dataSeries, index),
            name: dataSeries.label,
            barSize: 20,
            fill: dataSeries.color
          }
        );
      case "area":
        return /* @__PURE__ */ React.createElement(
          Area,
          {
            key,
            type: "monotone",
            data: normalizedData,
            dataKey: dataSeries.yKey,
            name: dataSeries.label,
            stroke: dataSeries.color,
            fill: dataSeries.color
          }
        );
      case "scatter":
        return /* @__PURE__ */ React.createElement(
          Scatter,
          {
            key,
            data: normalizedData,
            dataKey: dataSeries.yKey,
            name: dataSeries.label,
            fill: dataSeries.color
          }
        );
      default:
        return /* @__PURE__ */ React.createElement(
          Line,
          {
            key,
            type: "monotone",
            data: normalizedData,
            dataKey: dataSeries.yKey,
            name: dataSeries.label,
            stroke: dataSeries.color
          }
        );
    }
  })) : null)), /* @__PURE__ */ React.createElement(ErrorOverlay, { error: displayError }), loading && !error ? /* @__PURE__ */ React.createElement(
    Skeleton,
    {
      sx: { position: "absolute", inset: "0 0 0 0" },
      variant: "rectangular",
      width: "100%",
      height
    }
  ) : null);
}
var Chart_default = createComponent(Chart, {
  loadingProp: "loading",
  loadingPropSource: ["data"],
  errorProp: "error",
  resizableHeightProp: "height",
  argTypes: {
    data: {
      helperText: "The data to be displayed.",
      type: "array",
      schema: {
        type: "array",
        items: {
          type: "object",
          properties: {
            kind: {
              type: "string",
              enum: CHART_DATA_SERIES_KINDS,
              default: "line"
            },
            label: {
              type: "string"
            },
            data: {
              type: "object",
              default: []
            },
            xKey: {
              type: "string"
            },
            yKey: {
              type: "string"
            },
            color: {
              type: "string"
            }
          }
        }
      },
      control: { type: "ChartData", bindable: false }
    },
    height: {
      type: "number",
      default: 300,
      minimum: 100
    },
    sx: {
      helperText: SX_PROP_HELPER_TEXT,
      type: "object"
    }
  }
});

export {
  CHART_DATA_SERIES_KINDS,
  Chart_default
};
//# sourceMappingURL=chunk-X6EVMPMF.js.map